import inspect
import re
from functools import wraps

from aiozipkin.span import Span
import aiozipkin as az

from typing import (
    TYPE_CHECKING,
    Callable,
)

from tracuni.misc.helper import async_decorator
from tracuni.define.type import (
    SpanSide,
    UNKNOWN_NAME,
    Stage,
)
import tracuni.define.errors as err

if TYPE_CHECKING:
    from tracuni.point_accessor import PointAccessor
    from tracuni.point_accessor import MethodNameToStage
    from typing import Dict


# Отношение наименований методов аксессора точки к запускаемым фазам
# извлечения, отрабатывающим до выполнения этих методов.
method_to_stage = {
    # 'run_init_stage': Stage.INIT,
    # 'run_pre_stage': Stage.PRE,
    # 'run_post_stage': Stage.POST,
}  # type: MethodNameToStage

stage_to_method = {
}  # type: Dict[Stage, str]


def provide_span(stage: Stage, *args, **kwargs):
    """Декоратор для методов этапа
        * Гарантирует наличие экземпляра участка
        * Получает наполнение для текущей стадии
        * Сохранят результат вызова точки

    Parameters
    ----------
    stage
        Привязывает декорированный метод к конкретному этапу. Если таких
        несколько, то привязывается только первый обработанный

    Returns
    -------
        Метод обработки определённого этапа

    """
    def wrapper(fn):
        method_name = fn.__name__
        if method_name and stage not in method_to_stage:
            method_to_stage[method_name] = stage
            stage_to_method[stage] = method_name
        # Сохраняем прямое и обраное отображение этапов на методы обработки
        @wraps(fn)
        @async_decorator.coroutine
        def decorated(self, *args, **kwargs):
            # Определяем текущий этап
            stage_found = method_to_stage.get(method_name)
            if not stage_found:
                raise err.SpanNoStageForMethod

            # Для этапа после вызова точки сохраняем результат её работы
            if stage_found == Stage.POST:
                result = None
                if len(args):
                    result = args[0]
                if (
                    not result
                    and
                    hasattr(self.point.context.point_args, 'self')
                    and
                    hasattr(self.point.context.point_args.self,
                            'response_body')
                ):
                    result = self.point.context.point_args.self.response_body
                self.point.update_context({"point_result": result})

            # Вызываем извлечение и обработку данных этапа
            point_context = yield from self.point.register_stage(stage_found)
            self.point.context = point_context

            # отмечаем, если произошла ошибка
            error = point_context.reuse.get('error')
            if error:
                self.error = str(error)


            # обрабатываем выставленный обработчиками флаг пропуска
            if point_context.reuse.get('should_not_trace'):
                raise err.HTTPPathSkip

            # создаём участок, если он ещё не создан
            span_inst: Span = getattr(self, 'span_instance', None)
            if span_inst is None:
                if self.side == SpanSide.IN:
                    self.span_instance = self._fab_in_span()
                elif self.side == SpanSide.OUT:
                    self.span_instance = self._fab_out_span()
                span_inst = getattr(self, 'span_instance', None)
            if span_inst is None:
                raise err.NoSpanException

            if not self.point.context.span:
                self.point.update_context({"span": self})

            res = fn(self, span_inst, *args, **kwargs)
            return res

        return decorated

    return wrapper


class SpanGeneral:
    """Общая функциональность передачи данных на трассер

        * Реализует логику создания разных вариантов участков:
            * Входящий / Исходящий
            * С родительским контекстом / Без него
        * Записывает подготовленные данные в участок
            * В наименование
            * В заголовок
            * В аннотации

    """
    def __init__(self, side: SpanSide, point: 'PointAccessor'):
        self.span_instance = None
        self.side = side
        self.point = point
        self.tracer: az.Tracer = point.context.adapter.tracer
        self.error = None
        self.has_been_started = False

###############################################################################
# Создание участков для каждого из двух направлений
# Задействуются декоратором provide_span

    def _fab_in_span(self):
        # если в перехватываемых данных есть загловки трасера,
        # создаем дочерний входящий участок,
        # иначе создаем корневой входящий участок
        headers = self.point.context.headers
        context = az.make_context(headers)
        span_instance: Span = (
            self.tracer.new_trace()
            if context is None else
            self.tracer.new_child(context)
        )
        span_instance.kind(az.SERVER)
        # сохраняем идентификаторы входящего участка
        # все исходящие участки в данном дереве асинхронных вызовов
        # будут использовать его в качестве родительского контекста
        self.point.context.adapter.keep_context.set(
            'span',
            span_instance.context.make_headers(),
        )

        return span_instance

    def _fab_out_span(self):
        # если в контексте нет входящего участка,
        # то исходящая точка вызывается вне сценария трассровки,
        # возвращаемся в перехватчик и делаем прямой вызов без трассировки
        in_span_context = self.point.context.adapter.keep_context.get('span')
        if in_span_context is None:
            raise err.SpanNoParentException
        # исходящий участок всегда создаётся как дочерний к входящему
        in_span_context = az.make_context(in_span_context)
        span_instance = self.tracer.new_child(in_span_context)
        span_instance.kind(az.CLIENT)

        return span_instance

###############################################################################
# Методы прохождения по этапу

    @provide_span(Stage.INIT)
    def run_init_stage(self, span: Span):
        self.has_been_started = True
        span.start()
        self._load()

    @provide_span(Stage.PRE)
    def run_pre_stage(self, _: Span) -> Callable:
        if not self.has_been_started:
            raise err.SpanIsNotStartedException
        self._load()

        return self.point.apply_wrapped_fn()

    @provide_span(Stage.POST)
    def run_post_stage(self, span: Span, result):
        if not self.has_been_started:
            raise err.SpanIsNotStartedException
        self._load()
        span.finish(exception=self.error)

        return result

    def get_stage_method(self, stage: Stage):
        return getattr(self, stage_to_method.get(stage), lambda _: None)

    def register_error(self, e: Exception):
        self.error = e

    def enrich_headers(self, headers: dict, prefix_key=None):
        tracer_headers = self.span_instance.context.make_headers()
        if prefix_key:
            tracer_headers = {
                prefix_key: tracer_headers
            }
        headers.update(tracer_headers)
        return headers

###############################################################################
# Area privata
    def _load(self):
        span = self.span_instance
        span_name = self.point.context.span_name
        span_tags = self.point.context.span_tags
        span_logs = self.point.context.span_logs

        span.name(span_name.get('name', UNKNOWN_NAME))
        custom_remote = self.point.context.reuse['custom_remote_endpoint_name']
        standard_remote = span_name.get('remote_endpoint', UNKNOWN_NAME)
        remote = (
            custom_remote.format(**{'standard': standard_remote})
            if custom_remote else
            standard_remote
        )
        span.remote_endpoint(remote)

        if span_logs is None:
            span_logs = []
        if span_tags is None:
            span_tags = {}

        for log_item in span_logs:
            if log_item:
                self.span_instance.annotate(log_item)
        self.point.update_context({"span_logs": []})
        for tag_name, tag_value in span_tags.items():
            if tag_value:
                self.span_instance.tag(tag_name, tag_value)



import json

import aiohttp
import logging
import os
import ssl
import sys
import traceback
from urllib.parse import urlparse

from aiohttp import TCPConnector, ClientSession, client_exceptions

from app.core.component import Component
from app.core.helper import annotate_bytes


class HttpClient(Component):
    async def prepare(self):
        pass

    async def start(self):
        pass

    async def stop(self):
        pass

    async def request(
        self,
        method,
        url,
        body=None,
        time_out=None,
        headers=None,
        cert=None,
        **kwargs
    ):
        if cert:
            pem_file = '%s/cert/%s' % (os.path.realpath(os.path.dirname(sys.argv[0])), cert)
            sslcontext = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            sslcontext.load_cert_chain(pem_file)
        else:
            sslcontext = None

        if not time_out:
            time_out = self.app.config['system']['time_out']

        conn = TCPConnector(ssl_context=sslcontext)
        headers = headers or {}

        async with ClientSession(
                loop=self.loop,
                headers=headers,
                connector=conn
        ) as session:
            try:
                timeout = aiohttp.ClientTimeout(total=time_out)
                resp = await session._request(method, url, data=body, headers=headers, timeout=timeout, **kwargs)
            except BaseException as e:
                logging.error('[x] Error connecting to %s' % url)
                a_resp = None
            else:
                response_body = await resp.read()
                a_resp = await self.adapt_resp(resp)
                resp.release()
            return a_resp

    @staticmethod
    async def adapt_resp(resp):
        html = await resp.read()
        html = html.decode()
        http_code = resp.status
        _resp_hdrs = resp.headers
        resp_hdrs = {}
        for i in _resp_hdrs:
            resp_hdrs[i] = _resp_hdrs[i]

        # return {
        #     'http_code': http_code,
        #     'response_headers': resp_hdrs,
        #     'response': json.loads(html),
        # }
        return json.loads(html)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Перехват вызова интересующей точки
    Во время перехвата запускается процесс сбора, обработки и записи данных,
    согласно связаным с перехватчиком набором правил
"""
import time
import inspect
import logging
from functools import wraps
from typing import (
    Optional,
    Callable,
)

from tracuni.adapter import TracerWrapperRef
from tracuni.schema.engine.engine import SchemaEngine
from tracuni.span import (
    SpanGeneral,
)
from tracuni.point_accessor import (
    PointAccessor,
    PointContext,
)
from tracuni.define.const import (
    PATH_IGNORE_METHODS,
    PATH_IGNORE_PIECE,
    UNKNOWN_NAME,
    COLLECT_TRACE,
)
from tracuni.define.type import (
    Variant,
    RuleSet,
    Stage,
)
import tracuni.define.errors as err
from tracuni.misc.helper import (
    async_decorator,
    compose_own_call_stack,
    PostgreSQLError,
)


def tracer_point(
        variant: Variant,
        endpoint_name: str = None,
        point_ruleset: RuleSet = None,
):
    """

    Parameters
    ----------
    variant
        Назначить вариант обработки точки
    endpoint_name
        Здесь можно передать пользовательское имя удаленной точки
        при этом можно использовать и стандартное, указав в строке место
        подстановки с именем standard: 'mine: {standard}'
    point_ruleset
        Здесь можно передать свой вариант настройки извлечения формат такой
        же как значение одного из ключей стандартной настройки она заменит
        стандартную настройку конкретного варианта для данной точки,
        если же последним элементом пользовательской настройки будет строка
        '...', то пользовательская объединится со стандартной с приоритетом
        пользовательской в любом случае далее удут применены менее
        специфичные и менее приоритетные настройки, относящиеся к данной точке

    Returns
    -------
        Декорированный метод, вызов которого будет трассироваться

    """
    # создание экземпляра обработчика
    schema_engine = SchemaEngine(variant, point_ruleset)
    # подписываемся, чтобы знать состояние трассера
    TracerWrapperRef.register_state_observer(schema_engine)

    def wrapper(point_coro):
        @wraps(point_coro)
        @async_decorator.coroutine
        def decorated(*args, **kwargs):
            # сразу запускаем обернутую точку, если трассер выключен,
            # не соединился или не создан
            if schema_engine.is_tracer_disabled():
                result = yield from point_coro(*args, **kwargs)
                return result
            debug_is_on = schema_engine.adapter.config.debug

            stack_frames = (
                # первым элементом - название вызываемого метода
                {
                    'full_path': point_coro.__qualname__,
                    'short_path': point_coro.__name__,
                },
            )
            if COLLECT_TRACE:
                # собрать данные по стэку вызовов в данной точке
                # inspect.getouterframes(inspect.currentframe())

                frame_list = []
                frame = inspect.currentframe()
                while frame:
                    frame_list.append(frame)
                    frame = frame.f_back
                stack_frames = (
                    # добавить название вызываемого метода
                    *stack_frames,
                    *compose_own_call_stack(
                        PATH_IGNORE_METHODS,  # пропустить методы с именами
                        PATH_IGNORE_PIECE,  # пропустить модули по такому пути
                        name_dummy=UNKNOWN_NAME,  # не найденные имена
                        result_length=5,  # общий размер списка не превышает
                        from_index=1,  # начинать просмотр стэка с позиции
                        look_up_stop_idx=10,  # не просматривать дальше позиции
                        frame_list=frame_list,  # стэк
                    )
                )

            result = {}
            point_fn = None
            this_span: Optional[SpanGeneral] = None

            try:
                point_context = PointContext(
                    point_ref=point_coro,
                    adapter=schema_engine.adapter,
                    engine=schema_engine,
                    call_stack=stack_frames if stack_frames else {},
                )
                point = PointAccessor(
                    args,
                    kwargs,
                    point_context,
                    endpoint_name,
                )
                # создаем спан сконфигрированный по варианту и связанный с
                # ним и точкой
                this_span = point.fab_span_general()
                # если не создан участок переходим к прямому вызову
                if this_span:
                    yield from this_span.get_stage_method(Stage.INIT)()
                    # создаем участок, заполняем его данными доступными до
                    # вызова получаем готовый обернутый метод для вызова точки
                    pre_method = this_span.get_stage_method(Stage.PRE)
                    point_fn = yield from pre_method()  # type: Callable

            except Exception as e:
                # если здесь вылетели, то рабочий участок создать не удалось
                # schema_engine.tracer_wrap.log_debug('Error in tracer: {
                # }'.format(e))
                this_span = None
                if debug_is_on:
                    if not isinstance(
                        e,
                        (err.HTTPPathSkip, err.SpanNoParentException)
                    ):
                        logging.exception(e)
                    else:
                        logging.debug(e)

            finally:
                # в любом случае должны вызвать обернутую точку
                if point_fn and this_span:
                    # если обертка точки создана вызваем через неё
                    try:
                        result = yield from point_fn()
                    except (Exception, PostgreSQLError) as e:
                        # фиксируем данные об исключении
                        # ProgrammingError - ошибки обработки SQL запроса
                        if debug_is_on:
                            logging.exception(e)
                        this_span.register_error(e)
                    finally:
                        # так или иначе пробуем завершить формирование
                        # участка трассера
                        try:
                            result = yield from this_span.get_stage_method(
                                Stage.POST
                            )(
                                result
                            )
                        except Exception as e:
                            if debug_is_on:
                                logging.exception(e)

                else:
                    # если обертки точки нет, трассер не используется,
                    # вызываем напрямую
                    result = yield from point_coro(*args, **kwargs)

            return result

        return decorated

    return wrapper



from .ext_http_in import ruleset as http_in_ruleset
from .ext_http_out import ruleset as http_out_ruleset
from .ext_db_out import ruleset as db_out_ruleset
from .ext_amqp_out import ruleset as amqp_out_ruleset

__all__ = (
    'http_in_ruleset',
    'http_out_ruleset',
    'db_out_ruleset',
    'amqp_out_ruleset',
)



import datetime
import decimal
import inspect
import json
import re
import os
import sys
from itertools import islice
from typing import Iterable
import types
import logging

from ..define.const import USE_TORNADO

tornado_available = False
HTTPServer = None
PeriodicCallback = None
IOLoop = None

if USE_TORNADO:
    tornado_available = True
    try:
        # noinspection PyPackageRequirements
        from tornado.ioloop import IOLoop, PeriodicCallback
        # noinspection PyPackageRequirements
        from tornado.httpserver import HTTPServer
    except ImportError:
        tornado_available = False


async_decorator = types
if tornado_available:
    try:
        # noinspection PyPackageRequirements
        from tornado import gen as async_decorator
    except ImportError:
        pass

try:
    from psycopg2 import ProgrammingError as PostgreSQLError
except ImportError:
    class PostgreSQLError(BaseException):
        pass


def is_iter_but_str(item):
    return not isinstance(item, str) and isinstance(item, Iterable)


def to_iter(item):
    if is_iter_but_str(item):
        return item
    else:
        return item,


def _json_encoder(obj):
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S.%f%z')
    if isinstance(obj, datetime.date):
        return obj.strftime('%Y-%m-%d')
    if isinstance(obj, datetime.time):
        return obj.strftime('%H:%M:%S.%f%z')
    if isinstance(obj, datetime.timedelta):
        return obj.total_seconds()
    if isinstance(obj, bytes):
        try:
            return obj.decode('UTF8')
        except Exception:
            return str(obj)
    return repr(obj)


def json_dumps_with_dt(data, indent=0):
    return json.dumps(data, default=_json_encoder, indent=indent)


def get_project_path():
    return os.path.abspath(os.path.dirname(sys.argv[0]))


relative_path = __file__.replace(get_project_path(), '')
# app_path = os.path.normpath(os.path.split(relative_path)[0]).split(os.sep)[1]
project_path_pattern = re.compile(
    r'^{}/'.format(
        get_project_path(),
        # app_path
    )
)


def compose_own_call_stack(
        skip_methods,
        skip_paths,
        result_length=1,
        look_up_stop_idx=27,
        from_index=0,
        name_dummy='<unknown>',
        frame_list=None,
):
    outer_frames = frame_list if frame_list is not None else inspect.stack()

    def is_project_outer_frame(item):
        try:
            file_name = item.f_code.co_filename
            return (
                project_path_pattern.match(file_name)
                and
                "site-packages" not in file_name
                and
                "lib/python" not in file_name
            )
        except (IndexError, AttributeError):
            return False

    def compose_frame_path_object(item):
        try:
            file_name = project_path_pattern.sub(
                '',
                item.f_code.co_filename,
                1,
            )
            method_name = item.f_code.co_name
        except (IndexError, AttributeError):
            return None
        if method_name in skip_methods:
            return None
        if skip_paths in file_name:
            return None
        return {
            'full_path': file_name + ':' + method_name,
            'short_path': method_name,
        }

    lookup_list = [
        el
        for el in outer_frames[from_index:look_up_stop_idx]
        if is_project_outer_frame(el)
    ]
    composed_data_list = [
        compose_frame_path_object(el)
        for el in lookup_list
    ]
    compacted_list = [
        el
        for el in composed_data_list
        if el is not None
    ]
    result_dicts = list(islice(compacted_list, result_length))

    return result_dicts


def dict_from_json(data):
    """
    Return dict from json-string or dict
    :param data: string or dict
    :return: dict
    """
    try:
        data_dict = json.loads(data)
    except ValueError as e:
        logging.warning('JSON parse error: %s Data: %s.', e, data)
        data_dict = data
    return data_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Поставщик схем

    Разбор схем и формироавние списка правил для конкретного обработчика схем
"""

from collections import OrderedDict
from typing import (
    TYPE_CHECKING,
    Tuple,
    Optional,
)

from ...define.type import (
    Variant,
    RuleSet,
    Rule,
    SpanSide,
    APIKind,
    Schema,
    Extractors,
    Stage,
    DestinationSection,
    PipeCommand,
    OriginSection,
)
from ..standard import (
    standard_schema as imported_standard_schema,
    stage_to_sections,
    secret_mask_method_name,
)

if TYPE_CHECKING:
    from typing import Dict  # noqa
    from typing import Sequence  # noqa
    from ...define.type import RuleSet  # noqa
    from ...define.type import Schema  # noqa
from typing import List  # noqa
from ...define.type import Destination  # noqa
from typing import MutableSequence  # noqa
from ...define.type import MethodNameToStage  # noqa


class SchemaEngineFeed:
    """Подготовка схем для использования конкретным обработчиком

        * Получает ссылки на основную и пользовательские схемы и на
        определённый вариант.
        * Разбирает все схемы, выделяет из них относящиеся к заданному
        варианту и раскладывает их по фазам.
    """

    def __init__(self,
                 variant: Variant,
                 point_ruleset: Optional[RuleSet] = None,
                 custom_schema: Optional[Schema] = None,
                 ):
        self.point_ruleset = point_ruleset or tuple()
        self.custom_schema = custom_schema or {}
        self.standard_schema = imported_standard_schema or {}
        self.stage_to_sections = stage_to_sections or {}
        self.skip_secret_destinations = (
            []
        )  # type: MutableSequence[Destination]
        self.secret_mask_method_name = secret_mask_method_name

        self.variant = variant
        # Варианты, которые собираются из схем всех уровней в контексте
        # текущего конкретного варианта
        self.suitable_variants = OrderedDict([
            ('concrete', self.variant),
            ('all_kinds', Variant(self.variant.span_side, APIKind.ALL)),
            ('all_sides', Variant(SpanSide.ALL, self.variant.api_kind)),
            ('all_all', Variant(SpanSide.ALL, APIKind.ALL)),
        ])

    def parse(self) -> Extractors:
        """Корневая функция подготовки правил извлечения

            Объединяет все применимые для выбранного варианта наборы правил.
            Убирает неиспользуемые в силу приоритета или подавляющей
            команды. Распределяет наборы правил по этапам. Обрабатывает
            применение правил-команд для пользовательской схемы и набора
            правил точки. Для стандартной схемы их обработка не предусмотрена.

            Использует атрибуты своего объекта, ссылающиеся на
            схемы:
                * стандартную standard_schema
                * пользовательскую user_schema
                * точки point_ruleset

        Returns
        -------
            Отображение этапов извлечения на набор правил используемых на
            этом этапе.

        """
        # Выцепляем команды пропуска нижних приоритетов и маскирования из
        # набора правил точки. Если хотя бы одна комана пропуска относится
        # ко всем назначениям, то не применяем пользовательскую и стандартную
        # схемы. Делаем это отдельным первым шагом, чтобы не обрабатывать
        # схемы, если они должны быть пропущены
        custom_schema = {}  # type: Schema
        custom_ruleset = []  # type: Sequence
        standard_schema = {}  # type: Schema
        standard_ruleset = []  # type: Sequence
        (
            point_ruleset,
            skip_below_from_point_ruleset,
            skip_dest_secret,
        ) = (
            self._should_skip_level_below(self.point_ruleset)
        )  # type: Tuple[RuleSet, List[Destination], List[Destination]]
        if not (
            Destination(DestinationSection.ALL, '')
            in
            skip_below_from_point_ruleset
        ):
            standard_schema = self.standard_schema
            custom_schema = self.custom_schema
        self.skip_secret_destinations += skip_dest_secret

        # Получаем из пользовательской схемы набор правил по текущему
        # варианту и убираем из него правила, которые должны быть
        # пропущены согласно командам, выцепленным на предыдущем шаге.
        # Выцепляем команды пропуска нижних приоритетов и маскирования из
        # пользовательской схемы. Если хотя бы одна комана пропуска относится
        # ко всем назначениям, то не применяем стандартную схему.
        if custom_schema:
            (
                custom_ruleset,
                skip_below_from_custom_schema,
                skip_dest_secret,
            ) = self._should_skip_level_below(
                self._compose_variants(
                    custom_schema,
                    skip_below_from_point_ruleset,
                )
            )
            if (
                Destination(DestinationSection.ALL, '')
                in
                skip_below_from_custom_schema
            ):
                standard_schema = {}
            self.skip_secret_destinations += skip_dest_secret

            # Получаем из стандартной схемы набор правил по текущему
            # варианту и убираем из него правила, которые должны быть
            # пропущены согласно командам, выцепленным на предыдущих шагах.
            standard_ruleset = self._compose_variants(
                standard_schema,
                (
                    skip_below_from_custom_schema
                    +
                    skip_below_from_point_ruleset
                ),
            )
        else:
            standard_ruleset = self._compose_variants(
                standard_schema,
                [],
            )


        # объединяем все наборы правил из набора точки и двух схем
        collected_schema_rules = (
            *point_ruleset,
            *custom_ruleset,
            *standard_ruleset,
        )  # type: Tuple[Rule, ...]

        # раскладывем схемы по фазам и убираем из них правила, перекрытые в
        # силу приоритета
        return self._distribute_by_stages(collected_schema_rules)

    def get_skip_secret(self) -> MutableSequence[Destination]:
        return self.skip_secret_destinations or []

    def get_secret_mask_method_name(self) -> str:
        return self.secret_mask_method_name or ''

###############################################################################
# Area privata

    @staticmethod
    def _should_skip_level_below(
            schema_to_check: Optional[RuleSet] = None
    ) -> Tuple[RuleSet, List[Destination], List[Destination]]:
        """Применение пользовательской схемы поверх стандартной

        Специальной командой схемы задаётся заменяется ли стандартная
        схема пользовательской или объединяется с ней. В случае объединения
        приоритетными всё равно остаются пользовательские элементы.

        Пример элемента схемы, присутствие которого в пользовательской схеме
        отключает использование стандартной:
        (
            destination=Destination(
                section=DestinationSection.ALL,
            ),
            origins=tuple(),
            pipeline=tuple(PipeCommand.SKIP_LEVELS_BELOW,),
        )

        Parameters
        ----------
            schema_to_check

        Returns
        -------
            * Схему, из которой выброшены применённые правила
            * Список необрабатываемых назначений схем нижестоящего приоритета
            * Список необрабатываемых назначений по маскированию
        """
        if schema_to_check is None:
            schema_to_check = tuple()

        new_schema = []
        skip_dest_below = []
        skip_dest_secret = []
        for schema_rule in schema_to_check:
            skip_below_seq = tuple(
                el for el in schema_rule.pipeline
                if el == PipeCommand.SKIP_LEVELS_BELOW
            )
            if skip_below_seq:
                skip_dest_below.append(schema_rule.destination)
            skip_secret_seq = tuple(
                el for el in schema_rule.pipeline
                if el == PipeCommand.SKIP_SECRET_MASK
            )
            if skip_secret_seq:
                skip_dest_secret.append(schema_rule.destination)
            other_seq = tuple(
                el for el in schema_rule.pipeline
                if el not in (
                    PipeCommand.SKIP_LEVELS_BELOW,
                    PipeCommand.SKIP_SECRET_MASK,
                )
            )
            if len(skip_below_seq):
                if len(schema_rule.pipeline) > 1:
                    # noinspection PyProtectedMember
                    upd_rule = Rule(**{
                        **schema_rule._asdict(),
                        **{"pipeline": other_seq}
                    })
                    new_schema.append(upd_rule)
            else:
                new_schema.append(schema_rule)

        if list(
            d for d in skip_dest_below
            if d.section == DestinationSection.ALL
        ):
            skip_dest_below = [Destination(DestinationSection.ALL, ""), ]
        if list(
            d for d in skip_dest_secret
            if d.section == DestinationSection.ALL
        ):
            skip_dest_secret = [Destination(DestinationSection.ALL, ""), ]

        return (
            tuple(new_schema),
            skip_dest_below,
            skip_dest_secret,
        )

    def _compose_variants(self,
                          schema_map: Schema,
                          skip_dest: List[Destination],
                          ) -> RuleSet:
        """Получение набора правил для всех, подходящих к текущему,
        вариантов. Отбрасываются правила определённых назначений.

            * собираем из указанных для варианта текущего участка,
            * плюс применяемые для всех с данным направлением,
            * плюс для всех с данным протоколом,
            * плюс для всех вообще

        Parameters
        ----------
            schema_map
                входящая схема стандартная или пользовательская
            skip_dest
                список назначений, правил с которыми не применяюся из этой
                схемы

        Returns
        -------
            Единый набор правил, применимый для текущего варианта,
            полученный из данной схемы
        """
        return tuple(
            rule
            for ruleset in [
                schema_map.get(variant)
                for k, variant in self.suitable_variants.items()
            ]
            if ruleset is not None
            for rule in ruleset
            if rule.destination not in skip_dest
        )

    @staticmethod
    def _detect_stage(rule: Rule) -> Stage:
        """Относит правило к той или иной фазе

            Учитывается либо явное указание, либо наличие указанных в
            настройках секций источников и назначения

        Parameters
        ----------
        rule
            Проверяемое правило

        Returns
        -------
            Идентификатор фазы извлечения
        """
        if rule.stage is not None:
            return rule.stage
        else:
            # если другие условия не выполняются, то это экстрактор
            # предварительного этапа
            works_on_stage = Stage.PRE

            # если хотя бы одна секция источника относится к пост-этапу,
            # то и весь экстрактор работает там
            src_sections = set(src.section for src in rule.origins)
            post_sections = set(stage_to_sections[Stage.POST])
            leftovers = src_sections & post_sections

            if (
                (
                    leftovers
                    and
                    any(isinstance(el, OriginSection) for el in leftovers)
                )
                or rule.destination.section in post_sections
            ):
                works_on_stage = Stage.POST

            # если хотя бы одна секция назначения относится к этапу
            # инициализации, когда мы исключили пост,
            # то и весь экстрактор работает на этапе инициализации
            else:
                init_sections = set(stage_to_sections[Stage.INIT])
                leftovers = src_sections & init_sections
                if (
                    (
                        leftovers
                        and any(
                            isinstance(el, DestinationSection)
                            for el in leftovers
                        )
                    )
                    or rule.destination.section in init_sections
                ):
                    works_on_stage = Stage.INIT

            return works_on_stage

    def _distribute_by_stages(self, merged: RuleSet) -> Extractors:
        """Раскладывает применяемые схемы по стадиям

            Создаётся отношение кодового имени фазы к списку применяемых
            правил. При этом гарантируется применение только первого
            встреченного правила для каждого из уникальных назначений в
            независимости от того в какую фазу оно попадает

        Parameters
        ----------
        merged
            Подготовленная схема извлечения

        Returns
        -------
            Отношение стадий к применяемым схемам
        """
        rules_to_be_applied = {}  # type: Dict[Destination, dict]

        for rule in merged:
            # пропускаем менее специфичные, чем уже зарегистрированный
            # экстрактор
            if rule.destination not in rules_to_be_applied:
                rules_to_be_applied[rule.destination] = {
                    'stage': self._detect_stage(rule),
                    'extractor': rule,
                }

        # раскладываем собранные экстракторы по ключам этапов
        extractors = {}  # type: Extractors
        for stage in Stage:
            extractors[stage] = tuple(
                v['extractor']
                for k, v in rules_to_be_applied.items()
                if v['stage'] == stage
            )

        return extractors



import json
import logging
from urllib.parse import parse_qs

import datetime

import decimal

import io
from aiohttp import BytesPayload


class Error(Exception):
    pass


class GracefulExit(SystemExit):
    pass


class PrepareError(Error):
    pass


class TaskFormatError(Error):
    pass


class ValidationResponseError(Error):
    pass


class WidgetResponseError(Error):
    pass


class UnknownTaskError(Error):
    pass


class BadTaskParamsError(Error):
    pass


class PayErrorException(Error):
    pass


def handler_decorator(func):
    async def wrapper(*args, **kwargs):
        res = await func(*args, **kwargs)
        res.__setattr__('function', func.__name__)

        request = args[1]
        body = await request.text()

        if body:
            params = parse_qs(body)
            if not params:
                params = json.loads(body)
        else:
            params = {}

        res.__setattr__('request_headers', request.headers)
        res.__setattr__('request', params)
        return res

    return wrapper


def flatten(iterable):
    """
    :param iterable: {'hdrs': 'trololo', 'http_code': 12, 'terr': [1, 2, 3], 'z': [1]}
    :return: {'hdrs': 'trololo', 'http_code': 12, 'terr': [1, 2, 3], 'z': 1}
    """
    res = {}
    for k, v in iterable.items():
        if isinstance(v, (tuple, list)) and len(v) == 1:  # len(v) <= 1 # exclude dict
            res[k] = v[0]
        else:
            res[k] = v
    return res


async def get_post_params(request):
    body = await request.text()
    try:
        params = parse_qs(body)
        params = flatten(params)
    except Exception as e:
        logging.error('Cannot parse request body: %s.', body)
        params = {}
    return params


async def get_json_params(request):
    """
    :param request: '{"method": "PaymentInit", "msisdn": "79268638461", "gid": "1"}'
    :return: dict {'method': 'PaymentInit', 'msisdn': '79268638461', 'gid': '1'}
    """
    body = await request.text()
    try:
        params = json.loads(body)
        params = flatten(params)
    except Exception as e:
        logging.error('Cannot parse request body: %s.', body)
        params = {}
    return params


async def get_get_params(request):
    """
    :param request: 'http://inplat:8093/mocker/hexml/paymentbalance?ACCOUNT=InPlat&PWD=InPlat&MSISDN=79295803749'
    :return: dict {'ACCOUNT': 'InPlat', 'PWD': 'InPlat', 'MSISDN': '79295803749'}
    """
    get_row = request.GET
    try:
        params = dict(get_row)
    except Exception as e:
        logging.error('Cannot parse request get string: %s.', get_row)
        params = ''
    return params


def gen_log_config(config):
    if 'system' in config:
        sysconfig = config['system']
        log_config = {
            'log_format': sysconfig.get('log_format', "%(levelname)-8s [%(asctime)s] %(message)s"),
            'log_file': sysconfig.get('log_file', "logs/mocker2.log"),
            'log_mode': sysconfig.get('log_mode', "STDOUT"),
            'log_level': sysconfig.get('log_level', "INFO"),
        }
        return log_config
    else:
        raise KeyError("'system' does not exists in config!")


def set_python_log(config):

    levels = {
        'CRITICAL': logging.CRITICAL,
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG,
        'NOTSET': logging.NOTSET,
    }

    log_level = levels[config.get('log_level', 'INFO').upper()]

    log_formatter = logging.Formatter(config['log_format'])
    root_logger = logging.getLogger()

    def log2file():
        file_handler = logging.FileHandler(config['log_file'])
        file_handler.setFormatter(log_formatter)
        file_handler.setLevel(log_level)
        root_logger.addHandler(file_handler)

    def log2stdout():
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(log_formatter)
        console_handler.setLevel(log_level)
        root_logger.addHandler(console_handler)

    log_mode = config.get('log_mode', 'stdout').lower()

    if log_mode == 'file':
        log2file()
    elif log_mode == 'stdout':
        log2stdout()
    else:
        log2file()
        log2stdout()

    return root_logger


def json_encoder(obj):
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S.%f%z')
    if isinstance(obj, datetime.date):
        return obj.strftime('%Y-%m-%d')
    if isinstance(obj, datetime.time):
        return obj.strftime('%H:%M:%S.%f%z')
    if isinstance(obj, datetime.timedelta):
        return obj.total_seconds()
    if isinstance(obj, bytes):
        try:
            return obj.decode('UTF8')
        except:
            return str(obj)
    return repr(obj)


def json_encode(data):
    return json.dumps(data, default=json_encoder)


def annotate_bytes(span, data):
    if isinstance(data, BytesPayload):
        pl = io.BytesIO()
        data.write(pl)
        data = pl.getvalue()
    try:
        data_str = data.decode("UTF8")
    except Exception:
        data_str = str(data)
    span.annotate(data_str or 'null')


def raise_graceful_exit():
    raise GracefulExit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Привязка наборов правил к вариантам

    variant => ruleset

    ruleset = (
        Rule(
            description='',
            destination=Destination(
                section=DestinationSection.SPAN_TAGS,
                name='some'
            ),
            stage=Stage.PRE,
            origins=(
                Origin(section=OriginSection.REUSE, getter='some'),
            ),
            pipeline=(lambda x: x,),
        ),
    )  # type: RuleSet
"""

from tracuni.schema.builtin import http_in_ruleset
from tracuni.schema.builtin import http_out_ruleset
from tracuni.schema.builtin import db_out_ruleset
from tracuni.schema.builtin import amqp_out_ruleset

from tracuni.define.type import (
    Variant,
    SpanSide,
    APIKind,
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Dict
    # noinspection PyUnresolvedReferences
    from ..define.type import RuleSet  # noqa
    from ..define.type import Schema  # noqa
    from ..define.type import StageToSections  # noqa
    from ..define.type import MethodNameToStage  # noqa


####################
# Подключение стандартных настроек экстракторов

# для варианта будет использован только один из наборов для определённого
# ключа назначения в зависимости от разновидности ключа экстракторы
# применяются в следующем приоритете сверху вниз (если применяется верхний,
# то нижний уже не применяется)
#  - элементы с конкретным вариантом применяются
# только для этого варианта
#  - элементы привязанные только к направлению применяются ко всем вариантам
#  с этим направлением
#  - элементы привязанные по виду API применяются ко всем вариантам с этим
#  видом
#  - непривязанные к вариантам элементы применяются ко всем вариантам
standard_schema = {
    Variant(SpanSide.IN, APIKind.HTTP): http_in_ruleset,
    Variant(SpanSide.OUT, APIKind.HTTP): http_out_ruleset,
    # Variant(SpanSide.IN, APIKind.AMQP): amqp_in,
    Variant(SpanSide.OUT, APIKind.AMQP): amqp_out_ruleset,
    Variant(SpanSide.OUT, APIKind.DB): db_out_ruleset,
}  # type: Schema


####################
# Внутренние настройки разбора и применения правил извлечения

# Наличие данных секций в правиле относит его к начальной (INIT) или
# завершающей (POST) фазе, если стадия в правиле не указана явно. Приоритетом
# обладает завершающая фаза.
# Без указания фазы правило будет работать перед обращением к точке перехвата
# (PRE).
stage_to_sections = {
    Stage.INIT: [
        DestinationSection.REUSE,
    ],
    Stage.POST: [
        OriginSection.POINT_RESULT,
    ],
}  # type: StageToSections

secret_mask_method_name = 'pipe_secret'



from functools import partial

from tracuni.define.type import (
    Rule,
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    Stage,
)
from tracuni.schema.builtin.pipe import (
    ext_http_out_headers,
    ext_tracer_out_point,
    ext_http_out_url,
    pipe_inject_headers,
    pipe_head,
    pipe_sep_string,
    pipe_service_name_from_config,
    pipe_catch_essential,
    pipe_mask_secret,
    pipe_dump,
    pipe_cut,
    ext_send_request_args, ext_api_kind)


ruleset = (
    Rule(
        description="Записать заголовки в запрос",
        destination=Destination(DestinationSection.POINT_ARGS, 'headers'),
        pipeline=(
            pipe_inject_headers,
        ),
        origins=(
            Origin(
                OriginSection.SPAN, None,
            ),
            Origin(
                OriginSection.POINT_ARGS, ext_http_out_headers,
            ),
            Origin(
                OriginSection.ADAPTER,
                lambda adapter: adapter.config.service_name,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Наименование партнерского сервиса сохранить для "
                    "переиспользования",
        destination=Destination(
            DestinationSection.REUSE,
            'peer_name',
          ),
        pipeline=(
            pipe_service_name_from_config,
        ),
        origins=(
            Origin(OriginSection.ADAPTER, 'url_service_map'),
            Origin(OriginSection.POINT_ARGS, ext_http_out_url),
        ),
        stage=Stage.INIT,
    ),
    Rule(
        description="Задать имя участку",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'name'
        ),
        pipeline=(
            partial(pipe_sep_string, **{'sep': ("::", " (", ")")}),
        ),
        origins=(
            Origin(
                OriginSection.ENGINE,
                ext_api_kind,
            ),
            Origin(
                OriginSection.REUSE,
                "peer_name",
            ),
            Origin(
                OriginSection.CALL_STACK,
                ext_tracer_out_point,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Наименование партнерского сервиса записать в участок",
        destination=Destination(
            DestinationSection.SPAN_NAME,
            'remote_endpoint',
          ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.REUSE, 'peer_name'),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Хост партнерского сервиса",
        destination=Destination(
            DestinationSection.SPAN_TAGS,
            'peer.url',
          ),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.POINT_ARGS, ext_http_out_url),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Вывести тело запроса в журнал участка "
                    "- получить первый элемент источников"
                    "- выцепить существенные значения для тегов в строки"
                    "- сериализовать"
                    "- обрезать"
                    "- маскировать"
                    "- закавычить/откавычить/экранировать/энкодить",
        destination=Destination(DestinationSection.SPAN_LOGS, 'request_body'),
        pipeline=(
            pipe_head,
            pipe_catch_essential,
            pipe_dump,
            pipe_cut,
            pipe_mask_secret,
        ),
        origins=(
            Origin(
                OriginSection.POINT_ARGS,
                ext_send_request_args,
            ),
        ),
        stage=Stage.PRE,
    ),
    Rule(
        description="Сохранить информацию об ответе",
        destination=Destination(DestinationSection.REUSE, 'response'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(
                OriginSection.POINT_RESULT, None
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести тело ответа в аннотацию",
        destination=Destination(DestinationSection.SPAN_LOGS, 'response_body'),
        pipeline=(
            pipe_head,
            lambda v: v.get('response_body'),
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "response"
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести код ответа в аннотацию",
        destination=Destination(DestinationSection.SPAN_TAGS,
                                'http.rsp.status'),
        pipeline=(
            pipe_head,
            lambda v: v.get('response_code'),
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "response"
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести заголовки ответа в аннотацию",
        destination=Destination(DestinationSection.SPAN_TAGS,
                                'http.rsp.headers'),
        pipeline=(
            pipe_head,
            lambda v: v.get('response_headers'),
            pipe_dump,
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "response"
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести размер тела ответа в метки",
        destination=Destination(DestinationSection.SPAN_TAGS,
                                'http.rsp.size'),
        pipeline=(
            pipe_head,
            lambda v: len(v.get('response_body')),
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "response"
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Вывести сертификат ответа в метки",
        destination=Destination(DestinationSection.SPAN_TAGS,
                                'http.req.cert'),
        pipeline=(
            pipe_head,
            lambda v: v.get('request_cert'),
        ),
        origins=(
            Origin(
                OriginSection.REUSE, "response"
            ),
        ),
        stage=Stage.POST,
    ),
    Rule(
        description="Запомнить ошибку для вывода",
        destination=Destination(DestinationSection.REUSE, 'error'),
        pipeline=(
            pipe_head,
        ),
        origins=(
            Origin(OriginSection.SPAN, 'error'),
        ),
        stage=Stage.POST,
    ),
)



import asyncio

from app.core.app import BaseApp
from app.core.http_client import HttpClient
from app.core.http_server import HttpServer
from app.handlers.http_router import HttpRouter
from app.core.component import Component
from app.core.amqp_server import AMQPServer
from app.handlers.amqp_handler import AmqpHandler
from app.models.tasks_db import DbTasks


from tracuni import *


class Application(BaseApp):
    def __init__(self, config: dict, loop: asyncio.AbstractEventLoop):
        super(Application, self).__init__(config=config, loop=loop)
        self.attach_component(
            'db_tasks',
            DbTasks(config['db']),
            stop_after=['http_srv']
        )
        self.attach_component(
            "http_cln",
            HttpClient()
        )
        self.attach_component(
            "http_srv",
            HttpServer(
                self,
                config["system"]["host"],
                config["system"]["port"],
                HttpRouter,
                skip_trace=config["logging"]["skip_trace"]
            )
        )
        self.attach_component(
            'amqp_srv',
            AMQPServer(
                self,
                config['rabbit'],
                AmqpHandler
            )
        )
        rule_set = (
            Rule(
                destination=Destination(
                    section=DestinationSection.SPAN_TAGS,
                    name="some",
                ),
                pipeline=(
                    lambda _: "some",
                ),
                origins=(
                    Origin(
                        section=OriginSection.REUSE,
                        getter="custom_remote_endpoint_name",
                    ),
                ),
                stage=Stage.PRE,
            ),
            # Rule(
            #     destination=Destination(
            #         section=DestinationSection.ALL,
            #         name="",
            #     ),
            #     pipeline=(PipeCommand.SKIP_LEVELS_BELOW,),
            #     origins=tuple(),
            # )
        )  # type: RuleSet
        self.attach_component(
            'tracer_uni',
            fab_tracer(Component)(
                self,
                config['logging'],
                config['statsd'],
                # config_system=config['system'],
                # custom_schema={
                #     Variant(
                #         span_side=SpanSide.IN,
                #         api_kind=APIKind.HTTP,
                #     ): rule_set,
                # }  # type: Schema
            )
        )



# -*- coding: utf-8 -*-

"""Main module."""
import sys
import click

from typing import Any


def cube_primitive_test(x: int) -> int:
    return x * x


@click.command()
def main(args: Any = None) -> int:
    """Console script for trac_uni."""
    del args

    result_code = 0
    click.echo("Demo app for trac uni package")
    click.echo("2 x 2 = {}".format(cube_primitive_test(2)))
    return result_code


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from typing import (
    Sequence,
    Union,
    Any,
)


def do(
    seq: Sequence,
    indexes: Union[int, Sequence[int]]
) -> Any:
    is_single_value = isinstance(indexes, int)
    if is_single_value:
        indexes = (indexes,)
    reduced = []
    for idx in indexes:
        try:
            reduced.append(seq[idx])
        except IndexError:
            pass
    if not reduced:
        return
    if is_single_value:
        return reduced[0]
    return reduced



class ProcessorInitNoneOrWrongTypeVariant(Exception):
    """Variant is not found in provided VariantSchemaMaster
    dependencies"""


class ProcessorInitWrongTypeCustomSchema(Exception):
    """Provided Custom Schema is of wrong type"""


class ProcessorExtractCantSetDestinationAttr(Exception):
    """Invalid destination"""


class NoSpanException(Exception):
    """Attempt to process span without span itself, skipping"""


class SpanIsNotStartedException(Exception):
    """Attempt to fill in or finish span that has not been started"""


class SpanNoParentException(Exception):
    """Outgoing point without parent span,
        it's no error just a normal case flow
    """


class SpanNoStageForMethod(Exception):
    """Method decorated as stage-bound has no stage linked
    """


class HTTPPathSkip(Exception):
    """Skip HTTP path, it's no error just an option"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    # 'Click>=6.0',
    'typing;python_version<"3.6"',
    'aiozipkin>=0.4.0',
    'aiotask_context',
    "python-statsd",
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Konstantin Gonciarou",
    author_email='konstantin.goncharov@inplat.ru',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: Russian',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Convenient library to plug in tracer to incoming and "
                "outgoing points and extract data from there",
    entry_points={
        'console_scripts': [
            'tracuni=tracuni.cli:main',
        ],
    },
    install_requires=requirements,
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='jaeger zipkin tracer',
    name='tracuni',
    packages=find_packages(),
    python_requires='>=3.5.3',
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/hastly/trac_uni',
    version='0.2.0',
    zip_safe=False,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Узел подключения к другим компонентам
        * обеспечивает взаимодействие с главным компонентом клиентского
        приложения и инициализируется вмете с ним;
        * хранит ссылку на главный компонент клиентского приложения;
        * получает, обрабатывает и сохраняет настройки журналирования;
        * создаёт драйверы подключения к внешним ресурсам журналирования;
        * выбирает варианты реализации асинхронности (для работы с aiohttp
         и для работы с Tornado);
        * уведомляет остальные компоненты библиотеки о своей готовности;

        .side_steps.stat.intercept.TracerTransport
            параллельная запись метрик
        .misc.helper.async_decorator,
            для подключения к Tornado
                    .HTTPServer,
                    .IOLoop,
                    .tornado_available,
        .misc.keep_tornado.context_accessor,
            сохранение родительского участка, используя механизм Tornado
"""
import re
from abc import (
    ABC,
    abstractmethod,
)
from functools import partial
import logging

# драйвер трассера
import aiozipkin as az
import aiozipkin.transport as azt
# обертка драйвера сервиса метрик
from .side_steps.stat.intercept import TracerTransport as StatsTracerTransport

from .define.type import (
    TracerOptions,
    StatsdOptions,
)
from .define.const import (
    TRACER_SYSTEM_NAME,
)
from .misc.helper import (
    async_decorator,
    HTTPServer,
    IOLoop,
    tornado_available,
)

try:
    import aiotask_context
except ImportError:
    aiotask_context = None

if tornado_available:
    from .misc.keep_tornado import context_accessor


def init_tornado_loop(cls, port, app_conf, start_loop=False):
    # use only if tornado lib is available and tornado option in scheme
    # is on
    if not tornado_available:
        return
    logging_conf = app_conf.get('logging_conf')
    IOLoop.configure('tornado.platform.asyncio.AsyncIOLoop')
    this_app = cls(**app_conf)

    logging.info('%s listens at: %s' % (cls.__name__, port))

    tracer = fab_tracer()(
        this_app,
        logging_conf,
        app_conf.get('statsd_conf'),
        app_conf.get('system_conf'),
    )

    if tracer.config.enable:
        @async_decorator.coroutine
        def prehook(request):
            context_accessor.run_root_with_context(
                coro=partial(this_app, request)
            )

        HTTPServer(prehook).listen(port)
    else:
        HTTPServer(this_app).listen(port)

    tracer.run()
    if start_loop:
        IOLoop.current().start()

    return


class TracerObserve:
    """Временный класс подключаемый до полной инициализации адаптера

        Пока при запуске приложения ещё не создан полноценный класс делаем
        заглушку, которая позволяет другим объектам подписываться на
        изменения состояния трассера и получить уведомление о том,
        что к работе всё готово
    """
    observers = []

    @classmethod
    def register_state_observer(cls, observer):
        """Подписаться на уведомления о переходе в состоние готовности

        Parameters
        ----------
        observer
            экземпляр клиента с соответвующим методом, получающим уведомление
        """
        fn = getattr(observer, 'observe_state_change')
        if callable(fn):
            cls.observers.append(observer)


TracerWrapperRef = TracerObserve


def fab_tracer(component_class=None, ):
    if component_class is None:
        class TracerBase(ABC):
            variant = 'tornado'

            def __init__(self):
                super().__init__()
                self.tracer = None
                self.log_info = None
                self.log_debug = None
                self.client = None

            def run(self):
                self.log_info = getattr(self.client, 'log_info', logging.info)
                self.log_debug = getattr(self.client, 'log_debug', logging.debug)
                self.prepare_tracer()
                self.start_tracer()

            @abstractmethod
            def prepare_tracer(self):
                pass

            @abstractmethod
            def start_tracer(self):
                pass

    # адаптер для работы модуля трассера по интерфейсу компонента менджера
    # задач
    else:
        class TracerBase(component_class, ABC):
            variant = 'tm3'

            def __init__(self):
                super().__init__()
                self.tracer = None
                self.log_info = None
                self.log_debug = None

            async def prepare(self):
                self.log_info = getattr(self.client, 'log_info', logging.info)
                self.log_debug = getattr(self.client, 'log_debug', logging.debug)
                self.prepare_tracer()

            async def start(self):
                self.start_tracer()

            async def stop(self):
                await self.stop_tracer()

            async def stop_tracer(self):
                if self.tracer:
                    self.log_info("Shutting down tracer")
                    await self.tracer.close()
                    self.tracer = None
                    self._notify_observers()

            @abstractmethod
            def prepare_tracer(self):
                pass

            @abstractmethod
            def start_tracer(self):
                pass

    class TracerWrapper(TracerBase, TracerObserve):
        instance = None

        def __init__(self,
                     client,
                     config_logging,
                     config_statsd,
                     config_system=None,
                     custom_schema=None,
                     ):
            """

            Parameters
            ----------
            client
                ссылка на основной объект приложения клиента
            config_logging
                конфигурация трассера
            config_statsd
                конфигурация сервиса метрик
            config_system
                общая системная конфигуация, используется для получения
                строк имен сервис по адресам для именования исходящих участкоа
            custom_schema
                схема заменяющая или дополняющая стандартную схему бибилиотеки
            """
            super().__init__()
            self.custom_tracer_transport = None
            self.client = client
            self.custom_schema = custom_schema

            # заменяем фабрику задач для сохранения контекста, если не
            # используем Tornado
            if component_class is not None:
                self.loop = getattr(client, 'loop', None)
                self.loop.set_task_factory(
                    aiotask_context.copying_task_factory
                )
                self.keep_context = aiotask_context
            else:
                self.loop = IOLoop.current().asyncio_loop
                self.keep_context = context_accessor

            # устанавливаем настройки по умолчанию, потом пишем их поверх из
            # файла
            self.config: TracerOptions = TracerOptions()
            self.config = self._compose_options(config_logging, config_statsd)
            self.url_service_map = self._create_url_service_map(config_system)
            self.custom_tracer_transport = StatsTracerTransport(
                self.config,
                self.config.statsd,
                loop=self.loop,
            )

            skip_list_src = self.config.skip_trace
            self.skip_list = [
                re.compile(r'^{}$'.format(el))
                for el in skip_list_src
            ]

        def prepare_tracer(self):
            if TracerWrapper.instance is not None:
                return
            if not self.config.enable:
                return
            endpoint = az.create_endpoint(self.config.service_name)
            sampler = az.Sampler(sample_rate=self.config.sample_rate)

            transport = self.custom_tracer_transport
            if transport is None:
                transport = azt.Transport(
                    self.config.url,
                    send_interval=self.config.send_interval,
                    loop=self.loop
                )

            self.tracer = az.Tracer(transport, sampler, endpoint)

        def start_tracer(self):
            if self.tracer is not None:
                TracerWrapper.instance = self
                self._notify_observers()
                self.log_info('Tracer started to: {}'.format(
                    self.config.url))

        def should_not_trace(self, path):
            return any(el.match(path) for el in self.skip_list)

        def _notify_observers(self):
            for ob in self.observers:
                fn = getattr(ob, 'observe_state_change')
                if callable(fn):
                    fn(self)

        def _compose_options(self,
                             cfg_trace_log,
                             cfg_time_log
                             ) -> TracerOptions:
            # Включаем трассер,если имя в конфиге пристуствует и
            # соответствует заданному и если имя в конфиге или в значениях
            # по умолчанию не установлено в false
            unexpected_name = cfg_trace_log.get('tracer') != TRACER_SYSTEM_NAME
            disabled_by_default = self.config.enable
            disabled_in_config = cfg_trace_log.get(
                'tracer_enable',
                disabled_by_default
            ) is False
            tracer_enable = not (unexpected_name or disabled_in_config)

            # Получаем именнованный кортеж опций трассера с вложенным
            # кортежом опций statsd и сохранеными значениями по умолчанию у
            # того и другого, поскольку значения,
            # отсутствующие в файле конфигурации пропускаем
            read_statsd_opts = dict((k, v) for k, v in (
                ('enable', cfg_time_log.get('enable')),
                ('logging', cfg_time_log.get('logging')),
                ('url', cfg_time_log.get('url')),
                ('prefix', cfg_time_log.get('prefix')),
                ('hb_interval', cfg_time_log.get('hb_interval')),
            ) if v is not None)
            read_tracer_opts = dict((k, v) for k, v in (
                ('name', cfg_trace_log.get('tracer')),
                ('enable', tracer_enable),
                ('service_name', cfg_trace_log.get('tracer_svc_name')),
                ('url', cfg_trace_log.get('tracer_url')),
                ('sample_rate', cfg_trace_log.get('tracer_sample_rate')),
                ('send_interval', cfg_trace_log.get('tracer_send_interval')),
                ('skip_trace', cfg_trace_log.get('skip_trace')),
                ('debug', cfg_trace_log.get('debug')),
                ('statsd', StatsdOptions(**read_statsd_opts))
            ) if v is not None)
            return TracerOptions(**read_tracer_opts)

        @staticmethod
        def _create_url_service_map(system_conf):
            if system_conf is None:
                return {}
            return dict(
                (v, k)
                for k, v in system_conf.items()
                if isinstance(v, str)
            )

    global TracerWrapperRef
    TracerWrapperRef = TracerWrapper

    return TracerWrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Модуль доступа к вызываемой точке

"""
from collections import namedtuple
from copy import copy
from functools import partial
from inspect import signature, Parameter
from typing import (
    NamedTuple,
    Iterable,
    Callable,
    Optional,
    Sequence,
    Mapping,
    Dict,
    Union,
    Any,
    Tuple,
    Type,
)
from .define.type import (
    PointContext,
    MethodNameToStage,
    Stage,
)
from .misc.helper import async_decorator
from .span import SpanGeneral


class PointAccessor:
    """Доступ к точке вызова и связанным с ней данным

    * Хранит контекст: ссылки на точку, аргументы, сокращённый стек вызовов
    и схему варианта точки
    * Преобразует все аргументы к именованным, основываясь на сигнатуре
    точки и переданным именованным аргументам для упрощения доступа к ним
    * Копирует записываемое в трассер содержимое аргументов перед их
    модификацией
    * Регистрирует сигнал о достижении определённой фазы, вызывает извлечение
    данных этой фазы по схеме варианта
    * Хранит извлеченные данные в определённой структуре, предназначенной
    для записи в трассер и для переиспользования на последующих шагах работы
    экстрактора
    * Модифицирует ограниченное множество аргументов (заголовки HTTP)
    * Вызывает точку с оригинальными и модифицированными аргументами
    """

    def __init__(self,
                 original_args: Sequence,
                 original_kwargs: Dict,
                 context: PointContext,
                 custom_endpoint_name: str,
                 ):
        self._positional_names = []
        self._original_args = original_args
        self._original_kwargs = original_kwargs

        point_args = self._compose_kwargs(context.point_ref)

        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **context._asdict(),
            "point_args": point_args,
            "client": context.adapter.client,
            "reuse": {
                "custom_remote_endpoint_name": custom_endpoint_name,
                "error": None,
                "should_not_trace": False,
            },
            "headers": {},
            "span_name": {},
            "span_tags": {},
            "span_logs": [],
        })

    def fab_span_general(self):
        return SpanGeneral(
            self.context.engine.schema_feed.variant.span_side,
            self,
        )

    def apply_wrapped_fn(self) -> Callable:
        kwargs = copy(self._original_kwargs)
        kwargs.update(self.context.point_args._asdict())
        kwargs.pop('method_arguments_without_names')
        args = list(self._original_args)
        for idx, pos_name in enumerate(
            self._positional_names[:len(self._original_args)]
        ):
            args[idx] = kwargs.pop(pos_name, None)

        return partial(self.context.point_ref, *args, **kwargs)

    @async_decorator.coroutine
    def register_stage(self, stage_span_method_name: str) -> Any:
        res = yield from self.context.engine.extract(
            self.context,
            stage_span_method_name,
        )
        return res

    def update_context(self, new_data):
        # noinspection PyProtectedMember
        self.context = PointContext(**{
            **self.context._asdict(),
            **new_data,
        })

    def _compose_kwargs(self, point_ref: Callable) -> NamedTuple:
        sig = signature(point_ref).parameters
        self._positional_names = arg_names = [
            el
            for el in sig
            if el not in ('args', 'kwargs')
        ]
        kwargs = copy(self._original_kwargs)
        args = copy(self._original_args)
        kwargs.update(zip(arg_names, args))

        param_desc = sig
        for param in param_desc:
            pass_value = kwargs.get(param, param_desc.get(param).default)
            if pass_value != Parameter.empty:
                kwargs[param] = pass_value

        kwargs_l = len(kwargs)
        kwargs.update({'method_arguments_without_names': None})
        if kwargs_l < len(args):
            kwargs['method_arguments_without_names'] = args[kwargs_l:]

        point_args = namedtuple(
            'PointArguments',
            kwargs
        )(**kwargs)

        return point_args

    def __getitem__(self, code):
        # if not self.composed:
        #     raise PointWrapperError("Object should be populated with arguments before use")

        if all([
            isinstance(code, slice),
            *(getattr(code, el) is None for el in ('start', 'stop', 'step')),
        ]):
            return self._kwargs

        kwarg_setup = {}.get(code)
        if kwarg_setup is None:
            raise IndexError

        return self._kwargs.get(*kwarg_setup)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# noinspection PyProtectedMember

"""Проверка поставщика схем
"""

import types
from typing import Tuple

from tracuni.trac_uni import (
    Destination,
    DestinationSection,
    Origin,
    OriginSection,
    PipeCommand,
    Stage,
    PipelineTee,
    PointContext,
)
from tracuni.trac_uni import method_to_stage

import pytest


# pytest.skip('skipping {} tests'.format(__name__), allow_module_level=True)


@pytest.mark.asyncio
def test__skip_secret_use(
    engine_types,
    gen_schema_rule,
    variants,
    point_context,
):
    e_t, *_ = engine_types

    def test_pipe_secret(_):
        return "test"

    setup = (
        gen_schema_rule(
            destination=Destination(
                DestinationSection.ALL,
            ),
            pipeline=(PipeCommand.SKIP_LEVELS_BELOW,),
            stage=Stage.INIT,
        ),
        gen_schema_rule(
            origins=(Origin(OriginSection.POINT_ARGS, 'some'),),
            destination=Destination(
                DestinationSection.REUSE,
                "some",
            ),
            pipeline=(test_pipe_secret,),
            stage=Stage.INIT,
        ),
    )

    e = e_t(variants[0], point_ruleset=setup)
    init_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.INIT
    ][0]
    result = yield from e.extract(point_context, init_stage_method_name)
    assert result.reuse['some'] == 'test'


@pytest.mark.asyncio
def test__skip_secret_skip(
    engine_types,
    gen_schema_rule,
    variants,
    point_context,
):
    e_t, *_ = engine_types

    def test_pipe_secret(_):
        return "11"

    setup = (
        gen_schema_rule(
            destination=Destination(
                DestinationSection.ALL,
            ),
            pipeline=(PipeCommand.SKIP_LEVELS_BELOW,),
            stage=Stage.INIT,
        ),
        gen_schema_rule(
            origins=(Origin(OriginSection.POINT_ARGS, 'some'),),
            destination=Destination(
                DestinationSection.REUSE,
                "some",
            ),
            pipeline=(test_pipe_secret,),
            stage=Stage.INIT,
        ),
        gen_schema_rule(
            destination=Destination(
                DestinationSection.ALL,
            ),
            pipeline=(PipeCommand.SKIP_SECRET_MASK,)
        ),
    )

    e = e_t(variants[0], point_ruleset=setup)
    init_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.INIT
    ][0]
    result = yield from e.extract(point_context, init_stage_method_name)
    assert result.reuse['some'] != 'test'


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__should_skip_level_parse(
    engine_types,
    gen_schema_rule,
    custom_plus_standard_schema_map,
    variants,
    point_context,
):
    p_t, *_ = engine_types
    custom_schema = {
        variants[1]: (
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.REUSE,
                    'test_override',
                ),
                origins=(Origin(
                    OriginSection.POINT_ARGS,
                    lambda _: "custom",
                ),),
                pipeline=(lambda x: x[0],),
            ),
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.ALL,
                ),
                pipeline=(PipeCommand.SKIP_LEVELS_BELOW, lambda x: x,)
            ),
        ),
    }
    p = p_t(
        variants[1],
        custom_schema=custom_schema,
    )
    init_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.INIT
    ][0]
    point_context = yield from p.extract(point_context, init_stage_method_name)
    assert point_context.reuse['some'] == "value"
    assert point_context.reuse['test_override'] == "custom"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__skip_from_custom_map_all(
    engine_types,
    gen_schema_rule,
    custom_plus_standard_schema_map,
    variants,
    point_context,
):
    p_t, *_ = engine_types
    custom_schema_map = {
        variants[1]: (
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.REUSE,
                    'test_override',
                ),
                origins=(Origin(
                    OriginSection.POINT_ARGS,
                    lambda _: "custom",
                ),),
                pipeline=(lambda x: x[0],),
            ),
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.ALL,
                ),
                pipeline=(PipeCommand.SKIP_LEVELS_BELOW, lambda x: x,)
            ),
        ),
    }
    p = p_t(
        variants[1],
        custom_schema=custom_schema_map,
    )
    init_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.INIT
    ][0]
    point_context = yield from p.extract(point_context, init_stage_method_name)
    assert point_context.reuse['some'] == "value"
    assert point_context.reuse['test_override'] == "custom"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__skip_from_custom_map_specific(
    engine_types,
    gen_schema_rule,
    custom_plus_standard_schema_map,
    variants,
    point_context,
):
    p_t, *_ = engine_types
    custom_schema_map = {
        variants[1]: (
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.TAGS,
                    'test_override',
                ),
                origins=(Origin(
                    OriginSection.POINT_ARGS,
                    lambda _: "custom",
                ),),
                pipeline=(lambda x: x[0],),
            ),
            gen_schema_rule(
                destination=Destination(
                    DestinationSection.REUSE,
                    name="test_override",
                ),
                pipeline=(PipeCommand.SKIP_LEVELS_BELOW, lambda x: x,)
            ),
        ),
    }
    p = p_t(
        variants[1],
        custom_schema=custom_schema_map,
    )
    init_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.INIT
    ][0]
    point_context = yield from p.extract(point_context, init_stage_method_name)
    assert 'test_override' not in point_context.reuse
    assert point_context.tags['test_override'] == "custom"
    assert point_context.logs == ["standard"]


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__extract_values_miss(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    rule = gen_schema_rule(
        origins=(
            Origin(section=OriginSection.CLIENT, getter="some"),
            Origin(section=OriginSection.REUSE, getter="none"),
            Origin(section=OriginSection.CLIENT, getter="none"),
        ),
    )
    extracted = yield from simple_engine._extract_values(
        rule,
        point_context,
    )
    assert isinstance(extracted, tuple)
    assert len(extracted) == 0


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__extract_values_dict(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    rule = gen_schema_rule(
        origins=(Origin(section=OriginSection.REUSE, getter="some"),),
    )
    extracted = yield from simple_engine._extract_values(
        rule,
        point_context,
    )
    assert isinstance(extracted, tuple)
    assert len(extracted) > 0
    assert extracted[0] == "value"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__extract_values_obj(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    rule = gen_schema_rule(
        origins=(Origin(section=OriginSection.SPAN, getter="some"),),
    )
    extracted = yield from simple_engine._extract_values(
        rule,
        point_context
    )
    assert isinstance(extracted, tuple)
    assert len(extracted) > 0
    assert extracted[0] == "value"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__extract_values_coroutine(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    @types.coroutine
    def g(come):
        yield
        return come.get('some')

    rule = gen_schema_rule(
        origins=(Origin(section=OriginSection.REUSE, getter=g),),
    )
    extracted = yield from simple_engine._extract_values(
        rule,
        point_context,
    )
    assert isinstance(extracted, tuple)
    assert len(extracted) > 0
    assert extracted[0] == "value"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__extract_values_function(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    def g(come):
        return come.get('some')

    rule = gen_schema_rule(
        origins=(Origin(section=OriginSection.REUSE, getter=g),),
    )
    extracted = yield from simple_engine._extract_values(
        rule,
        point_context,
    )
    assert isinstance(extracted, tuple)
    assert len(extracted) > 0
    assert extracted[0] == "value"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test__extract_values_root(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    rule = gen_schema_rule(
        origins=(Origin(section=OriginSection.REUSE, getter="/"),),
    )
    extracted = yield from simple_engine._extract_values(
        rule,
        point_context,
    )
    assert isinstance(extracted, tuple)
    assert len(extracted) > 0
    assert isinstance(extracted[0], dict)
    assert extracted[0].get("some") == "value"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test__pipe_values_instance(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    v = simple_engine._pipe_values(
        gen_schema_rule(),
        point_context,
        1,
    )
    assert v.logs == [1]


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test__pipe_values_tee(
    simple_engine,
    gen_schema_rule,
    point_context,
):
    upd_point_context = {
        OriginSection.POINT_ARGS.name.lower(): "old_value",
        DestinationSection.LOGS.name.lower(): ["old_value"],
        DestinationSection.TAGS.name.lower(): {"a": "old_value"},
    }
    test_point_context = PointContext(**{
        **point_context._asdict(),
        **upd_point_context,
    })

    def pipe_tee(v):
        return PipelineTee(
            main_story=v,
            side_story={
                OriginSection.POINT_ARGS: v,
                DestinationSection.LOGS: [v],
            }
        )
    test_point_context = simple_engine._pipe_values(
        gen_schema_rule(
            destination=Destination(
                DestinationSection.TAGS,
                'b',
            ),
            pipeline=(pipe_tee,),
        ),
        test_point_context,
        "new_value",
    )
    assert test_point_context.point_args == "new_value"
    assert test_point_context.logs == ["new_value"]
    assert test_point_context.tags == {"a": "old_value", "b": "new_value"}


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
def test__update_point_context_empty(
    simple_engine,
    gen_schema_rule,
    point_context,
    engine_types,
):
    err = engine_types[-1]
    test_point_context = PointContext(**point_context._asdict())
    simple_engine._update_point_context(
        gen_schema_rule(destination=Destination(DestinationSection.ALL)),
        test_point_context,
        "nothing",
    )
    assert point_context == test_point_context

    test_point_context = PointContext(**{
        **point_context._asdict(),
        **{
            'logs': 1
        },
    })
    with pytest.raises(
        err.ProcessorExtractCantSetDestinationAttr,
    ):
        simple_engine._update_point_context(
            gen_schema_rule(
                destination=Destination(
                    section=DestinationSection.LOGS,
                    name='test'
                ),
            ),
            test_point_context,
            "nothing",
        )

    class Test:
        pass
    test_point_context = PointContext(**{
        **point_context._asdict(),
        **{
            'logs': Test()
        },
    })
    res = simple_engine._update_point_context(
        gen_schema_rule(
            destination=Destination(
                section=DestinationSection.LOGS,
                name="test"
            ),
        ),
        test_point_context,
        "some",
    )
    assert res.logs.test == "some"


# @pytest.mark.skip(reason="get out of the way of the best boosted beasts")
@pytest.mark.asyncio
def test_extract(
    engine_types,
    variants,
    gen_schema_rule,
    point_context,
):

    def pipe_rev(x: Tuple[str]):
        return ''.join(reversed(x[0]))

    p_t = engine_types[0]
    p = p_t(
        variants[0],
        (
            gen_schema_rule(
                'ignore other rules',
                destination=Destination(
                    section=DestinationSection.ALL
                ),
                pipeline=(PipeCommand.SKIP_LEVELS_BELOW,)
            ),
            gen_schema_rule(
                'collect logs',
                destination=Destination(
                    section=DestinationSection.LOGS,
                    name="first",
                ),
                origins=(
                    Origin(
                        section=OriginSection.POINT_ARGS,
                        getter='some',
                    ),
                ),
                pipeline=(pipe_rev,),
                stage=Stage.INIT,
            ),
            gen_schema_rule(
                'collect logs',
                destination=Destination(
                    section=DestinationSection.LOGS,
                    name="second",
                ),
                origins=(
                    Origin(
                        section=OriginSection.LOGS,
                        getter=lambda x: x[0],
                    ),
                ),
                pipeline=(pipe_rev,),
                stage=Stage.PRE,
            ),
            gen_schema_rule(
                'collect logs',
                destination=Destination(
                    section=DestinationSection.LOGS,
                    name="third",
                ),
                origins=(
                    Origin(
                        section=OriginSection.LOGS,
                        getter=lambda x: None,
                    ),
                ),
                pipeline=(lambda x: x,),
                stage=Stage.POST,
            ),
        ),

    )
    init_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.INIT
    ][0]
    pre_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.PRE
    ][0]
    post_stage_method_name = [
        k for k, v in method_to_stage.items()
        if v == Stage.POST
    ][0]
    res = yield from p.extract(point_context, init_stage_method_name)
    assert res.logs == ['eulav']
    res = yield from p.extract(res, pre_stage_method_name)
    assert res.logs == ['eulav', 'value']
    res = yield from p.extract(point_context, post_stage_method_name)
    assert res.logs == ['eulav', 'value']
    res = yield from p.extract(res, 'wrong')
    assert res.logs == ['eulav', 'value']



class Error(Exception):
    pass


class GracefulExit(SystemExit):
    pass


class PrepareError(Error):
    pass


class TaskFormatError(Error):
    pass


class ValidationResponseError(Error):
    pass


class WidgetResponseError(Error):
    pass


class UnknownTaskError(Error):
    pass


class BadTaskParamsError(Error):
    pass


class PayErrorException(Error):
    pass
